/*
 * miscellaneous routines for dpd and dplogin, mostly OS-dependent
 */

#include <stdio.h>
#include <malloc.h>
#include <signal.h>
#include <errno.h>
#include <setjmp.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/fcntl.h>
#include <sys/param.h>
#include <sys/stream.h>
#include <sys/resource.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <sys/socket.h>
#include <net/if.h>
#include <stropts.h>

#ifndef SOLARIS_1
#include <sys/dlpi.h>
#endif

#include "args.h"
#include "dpio.h"
#include "dp.h"
#include "dpd.h"
#include "ppp.h"

int	sys_set_timeouts	__ARGS((REMOTE *, dp_timeo_t *));
void	getpacketcounts		__ARGS((REMOTE *, dp_stats_t *));
void	if_callstat		__ARGS((REMOTE *, char));

#ifdef SOLARIS_1

int
sys_set_timeouts(rp, dpt)
REMOTE *rp;
dp_timeo_t *dpt;
{
    static char WHERE[] = "sys_set_timeouts";
    int s;
    struct ifreq ifr;

    if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	d_log(DLOG_GENERAL, WHERE,
	      "Can't create socket to configure timeouts for \"%s\", %m",
	      rp->Device);
	return -1;
    }
    (void)strcpy(ifr.ifr_name, rp->Device);
    ((u_long *)(ifr.ifr_data))[0] = dpt->dt_active;
    ((u_long *)(ifr.ifr_data))[1] = dpt->dt_active_c;
    ((u_long *)(ifr.ifr_data))[2] = dpt->dt_active_u;
    if (ioctl(s, SIOCSDPIATIMEO, (caddr_t)&ifr) < 0) {
	d_log(DLOG_GENERAL, WHERE,
	      "Can't set Dial Activity Timeouts on \"%s\", %m", rp->Device);
	(void)close(s);
	return -1;
    }
    (void)strcpy(ifr.ifr_name, rp->Device);
    ((u_long *)(ifr.ifr_data))[0] = dpt->dt_waiting;
    ((u_long *)(ifr.ifr_data))[1] = dpt->dt_failcall;
    if (ioctl(s, SIOCSDPICTIMEO, (caddr_t)&ifr) < 0) {
	d_log(DLOG_GENERAL, WHERE,
	      "Can't set Dial Call Timeouts on \"%s\", %m", rp->Device);
	(void)close(s);
	return -1;
    }
    (void)close(s);
    return 1;
}

/*
 *  Get the packet counts from the interface.
 */

#define	GETIFUINTAR(ifr, i)	(((int *)((ifr).ifr_data))[i])

void
getpacketcounts(rp, ds)
REMOTE *rp;
dp_stats_t *ds;
{
struct slipstat *ss;
    static char	WHERE[] = "getpacketcounts";
    struct ifreq ifr;
    int s;

    (void)memset((char *)&ifr, 0, sizeof(ifr));
    (void)strcpy(ifr.ifr_name, rp->Device);

    if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	d_log(DLOG_GENERAL, WHERE, "Can't create socket for \"%s\", %m",
	    rp->Device);
	unlock_pid();
	exit(1);
    }
    if (ioctl(s, SIOCGDPIISTATS, (caddr_t)&ifr) < 0) {
	d_log(DLOG_GENERAL, WHERE, "Can't SIOCGDPIISTATS for \"%s\", %m",
	      rp->Device);
	unlock_pid();
	exit(1);
    }

    ds->dps_ibytes   = GETIFUINTAR(ifr, 0);
    ds->dps_ipackets = GETIFUINTAR(ifr, 1);
    ds->dps_ierrors  = GETIFUINTAR(ifr, 2);

    if (ioctl(s, SIOCGDPIOSTATS, (caddr_t)&ifr) < 0) {
	d_log(DLOG_GENERAL, WHERE, "Can't SIOCGDPIOSTATS for \"%s\", %m",
	      rp->Device);
	unlock_pid();
	exit(1);
    }

    ds->dps_obytes   = GETIFUINTAR(ifr, 0);
    ds->dps_opackets = GETIFUINTAR(ifr, 1);
    ds->dps_oerrors  = GETIFUINTAR(ifr, 2);
    (void)close(s);
}

/*
**  Log that a call failed.
*/
void
if_callstat(rp, stat)
REMOTE *rp;
char stat;
{
    static char		WHERE[] = "if_callstat";
    int			s;
    struct ifreq	ifr;

    if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	d_log(DLOG_GENERAL, WHERE, "Can't create socket for \"%s\", %m",
	    rp->Device);
	return;
    }

    (void)strcpy(ifr.ifr_name, rp->Device);
    ifr.ifr_data[0] = stat;
    if (ioctl(s, SIOCICALLSTAT, (caddr_t)&ifr) < 0)
	d_log(DLOG_GENERAL, WHERE, "Can't SIOCICALLSTAT on \"%s\", %m",
	    rp->Device);
    (void)close(s);
}

#else /* SOLARIS_2 */

int
sys_set_timeouts(rp, dpt)
REMOTE *rp;
dp_timeo_t *dpt;
{
    return strioctl(rp->DlpiFd, DPIOCSTIMEOS, -1, sizeof(dp_timeo_t), (char *)dpt);
}

/*
 *  Get the packet counts from the interface.
 */

void
getpacketcounts(rp, ds)
REMOTE *rp;
dp_stats_t *ds;
{
    static char	WHERE[] = "getpacketcounts";
    int s;

    if (strioctl(rp->DlpiFd, DPIOCGIFSTATS, -1, sizeof(*ds), (char *)ds) < 0) {
	d_log(DLOG_GENERAL, WHERE, "Can't DPIOCGIFSTATS for \"%s\", %m",
	      rp->Device);
	unlock_pid();
	exit(1);
    }
}

/*
**  Log that a call failed.
*/
void
if_callstat(rp, stat)
REMOTE *rp;
char stat;
{
    static char		WHERE[] = "if_callstat";

    if (strioctl(rp->DlpiFd, DPIOCCALLSTAT, -1,
		 sizeof(stat), &stat) < 0) {
	d_log(DLOG_GENERAL, WHERE, "Can't DPIOCCALLSTAT on \"%s\", %m",
	      rp->Device);
    }
}

#endif /* SOLARIS_1 or SOLARIS_2 */
