#ifndef	_SYS_DPIO_H
#define	_SYS_DPIO_H

#ifdef	__cplusplus
extern "C" {
#endif

/*
 * Dialup PPP definitions to communicate to the kernel modules.
 *
 * Copyright (c) 1992 Purdue University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Purdue University.  The name of the University may not be used
 * to endorse or promote products derived * from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

typedef struct dp_timeo {
    int dt_active,
	dt_active_c,
	dt_active_u,
	dt_waiting,
	dt_failcall;
} dp_timeo_t;

#define	DPT_NTIMEOUTS		(sizeof(dp_timeo_t)/sizeof(int))

#define	DPT_NO_TIMEOUT		(-1)
#define	DPT_DEF_TIMEOUT		(-2)
#define	DPT_UNSPEC_TIMEOUT	(-3)

typedef struct dp_stats {
    int dps_ibytes,
	dps_ipackets,
	dps_ierrors;
    int dps_obytes,
	dps_opackets,
	dps_oerrors;
} dp_stats_t;


/*
 * Bits to OR into the control field of a PPP DLSAP. 
 *
 * DPC_DIAL_REQ		Route packets to this stream when outgoing packets
 *			matching this SAP cannot be sent because the
 *			modem is disconnected and needs dialing.
 * DPC_PROTO_RANGE	Match only on the upper 4 bits of the proto part
 *			of the SAP.  The PPP proto field is 16 bits and
 *			currently defines the following ranges of values:
 *			0XXX	Network Protocol Datagram (e.g. IP, etc)
 *			8XXX	Network Control Protocol (NCP) Datagram
 *			CXXX	Link Control Protocol (LCP) Datagram
 */

#define	DPP_DIAL_REQ	0x0100
#define	DPP_SNOOPED	0x0200
#define	DPP_PROTO_MATCH	0x0001
#define	DPP_PROTO_MASK	0xf000


/*
 * Definitions for constructing DLSAP addresses from their componentsT
 */
#define	DP_DLSAP(addr, ctl, proto)	((((unsigned long)(addr) &0xff)<<24) | \
					 (((unsigned long)(ctl)  &0xff)<<16) | \
					  ((unsigned long)(proto)&0xffff))
#define	DP_DLSAP_ADDR(sap)		(((sap) >> 24) & 0xff)
#define	DP_DLSAP_CTL(sap)		(((sap) >> 16) & 0xff)
#define	DP_DLSAP_PROTO(sap)		((sap) & 0xffff)

#define	DP_DIAL_SAP(sap)		((sap) |  DPP_DIAL_REQ)
#define	DP_PROTO_SAP(sap)		((sap) & ~DPP_PROTO_MATCH)

/*
 * Ioctl's that are recognized by the dp device.
 * These are all availble exclusively through the stream ioctl mechanism.
 */
#ifdef	SOLARIS_1
#include <sys/ioccom.h>
#ifdef	__STDC__
#define	_DPW(n, t)	_IOW('P', n, t)
#define	_DPR(n, t)	_IOR('P', n, t)
#else
#define	_DPW(n, t)	_IOW(P, n, t)
#define	_DPR(n, t)	_IOR(P, n, t)
#endif
#else
#define	DPIOC		('P' << 8)
#define	_DPW(n, t)	(DPIOC|(n))
#define	_DPR(n, t)	(DPIOC|(n))
#endif

#define	DPIOCIFATTACH	_DPW(0x01, long)	/* Attach network device */
#define	DPIOCIFDETACH	_DPW(0x02, long)	/* Detach network device */

#define	DPIOCSTIMEOS	_DPW(0x03, dp_timeo_t)	/* Set Timeouts */
#define	DPIOCGTIMEOS	_DPR(0x04, dp_timeo_t)	/* Get Timeouts */
#define	DPIOCCALLSTAT	_DPW(0x05, char)	/* Report Call Status */
#define	DPIOCSAPENABLE	_DPW(0x06, long)	/* Enable SAP output */
#define	DPIOCSAPDISABLE	_DPW(0x07, long)	/* Disable SAP output */

#define	DPIOCGIFSTATS	_DPR(0x08, dp_stats_t)	/* Retrieve statistics */

#define	DPIOCSIFP2PFL	_DPW(0x09, long)	/* Set P2P flag for IP */

#define	DPIOCSDPDEBUG	_DPW(0x0e, int)		/* Set Debug flag */
#define	DPIOCGDPDEBUG	_DPR(0x0f, int)		/* Get Debug flag */

/*
 * These ioctl's are particular to the Van Jacobson compression module.
 * The dp device will also pass these downstream to get them to a vjc module.
 */
#define	DPIOCSVJCOMP	_DPW(0x10, short)	/* Enable/disable compression */
#define	DPIOCSVJCDEBUG	_DPW(0x1e, int)		/* Set Debug flag */
#define	DPIOCGVJCDEBUG	_DPR(0x1f, int)		/* Get Debug flag */

/*
 * These ioctl's are particular to the PPP Async module.
 * The dp device will also pass these downstream to get them to the ppp_async
 * module.
 */
#define	DPIOCSCOMPAC	_DPW(0x20, char)	
#define	DPIOCSCOMPPROT	_DPW(0x21, char)
#define	DPIOCSMRU	_DPW(0x22, int)
#define	DPIOCGMRU	_DPR(0x23, int)
#define	DPIOCSASYNCMAP	_DPW(0x24, long)
#define	DPIOCGASYNCMAP	_DPR(0x25, long)
#define	DPIOCSFRAMING	_DPW(0x26, char)	/* Set framing method*/
#define	DPIOCGFRAMING	_DPR(0x27, char)	/* Get framing method*/
#define	DPIOCSASDEBUG	_DPW(0x2e, int)		/* Set Debug flag */
#define	DPIOCGASDEBUG	_DPR(0x2f, int)		/* Get Debug flag */

/*
 * These ioctl's are for the Solaris 1.X version
*/
#define	DPIOCSIFUNIT	_DPW(0x30, int)
#define	DPIOCGIFUNIT	_DPR(0x31, int)
/*
 * Supported framing methods.
 */
#define	DP_FRAME_PPP	0
#define	DP_FRAME_SLIP	1

/*
 * Call status values (for SIOCCALLSTAT)
 */
#define	DP_FAILURE	0	/* Failure to establish modem connection */
#define	DP_SUCCESS	1	/* Connection is established */
#define	DP_IN_PROGRESS	2	/* Connect is being established */
#define	DP_NO_MODEM	3	/* No modem is available */
#define DP_VALID_STATUS(s)	((s) >= 0 && (s) <= 3)

/*
 * Defines for control messages passed between modules.
 */
#define	DP_IF_IERROR	0xe1	/* pppasync -> dp or vjc -> dp */
#define	DP_IF_OERROR	0xe2	/* pppasync -> dp or vjc -> dp */

/*
 * Defines for DLPI messages to the DP device.
 */
#define	DP_MAXDLBUF	PPP_MTU
#define	DP_MAXWAIT	15

/*
 * Path to the DP device
 */
#define	DEVDP			"/dev/dp"

/*
 * A specially defined DLPI primitive to handle inactivity timeouts.
 * There should be a defined way to do this.
 */
typedef struct {
    unsigned long dl_primitive;
    unsigned long dl_signal;
} dl_signal_ind_t;

#define	DL_SIGNAL_IND		0x1001

#define	DL_SIGNAL_IND_SIZE	sizeof (dl_signal_ind_t)

#ifdef	SOLARIS_1
#ifdef	__STDC__
#define	_DPIW(n, t)	_IOW('i', n, t)
#define	_DPIR(n, t)	_IOR('i', n, t)
#define	_DPIWR(n, t)	_IOWR('i', n, t)
#else
#define	_DPIW(n, t)	_IOW(i, n, t)
#define	_DPIR(n, t)	_IOR(i, n, t)
#define	_DPIWR(n, t)	_IOWR(i, n, t)
#endif
/*
 * These ioctl's are necessary for Solaris 1 since they can be used
 * without having a unit open.
 */
#define SIOCICALLSTAT	_DPIW( 64, struct ifreq) /* Report Call Status */
#define	SIOCSDPIATIMEO	_DPIW( 65, struct ifreq) /* Set Activity Timeouts */
#define	SIOCGDPIATIMEO	_DPIWR(66, struct ifreq) /* Get Activity Timeouts */
#define	SIOCSDPICTIMEO	_DPIW( 67, struct ifreq) /* Set Call Timeouts */
#define	SIOCGDPICTIMEO	_DPIWR(68, struct ifreq) /* Get Call Timeouts */
#define	SIOCGDPIISTATS	_DPIWR(69, struct ifreq) /* Get Input Stats */
#define	SIOCGDPIOSTATS	_DPIWR(70, struct ifreq) /* Get Output Stats */

/*
 * stuff borrowed from Solaris 2 <sys/dlpi.h>
 */

typedef unsigned long ulong;

/*
 * DL_UNITDATA_IND, M_PROTO type, with M_DATA block(s)
 */
typedef struct {
	ulong	dl_primitive;		/* DL_UNITDATA_IND */
	ulong	dl_dest_addr_length;	/* DLSAP length of dest. user */
	ulong	dl_dest_addr_offset;	/* offset from beg. of block */
	ulong	dl_src_addr_length;	/* DLSAP addr length of sending user */
	ulong	dl_src_addr_offset;	/* offset from beg. of block */
	ulong	dl_group_address;	/* set to one if multicast/broadcast */
} dl_unitdata_ind_t;

#define	DL_UNITDATA_IND		0x08	/* datagram receive indication */

#endif

#define	DR_MINHDRLEN	(sizeof(struct ip))
#define	DR_MAXHDRLEN	(sizeof(struct ip) + 2 * sizeof(u_short))

typedef struct dp_reqinfo {
    char dri_ifname[8];
    int dri_ifunit;
    int dri_flag;
    struct sockaddr_in dri_sin;
    int dri_hdrlen;
} dp_reqinfo;

typedef union dp_reqhdr {
    char drh_data[DR_MAXHDRLEN];
    struct ip drh_ip;
} dp_reqhdr;

typedef struct dp_req {
    struct dp_reqinfo dr_info;
    union dp_reqhdr dr_hdr;
} dp_req;

#define	dr_ifname	dr_info.dri_ifname
#define	dr_ifunit	dr_info.dri_ifunit
#define	dr_flag		dr_info.dri_flag
#define	dr_sin		dr_info.dri_sin
#define	dr_hdrlen	dr_info.dri_hdrlen

#define	dr_data		dr_hdr.drh_data
#define	dr_ip		dr_hdr.drh_ip
#define	dr_tcpip	dr_hdr.drh_tcpip
#define	dr_udpip	dr_hdr.drh_udpip

#define DP_REQ_CHECKIT	1

#ifdef	__cplusplus
}
#endif

#endif	/* _SYS_DPIO_H */
