#include <sys/types.h>
 
#include "ppp.h"
 
#include <net/if.h>
#include <sys/sockio.h>
#include <net/route.h>
#include <syslog.h>
#include <string.h>
#include <unistd.h>
 
int strioctl __ARGS((int, int, int, int, char *));

#ifndef ifr_mtu
#define ifr_mtu		ifr_metric
#endif

#if 0
/*
 * ppp_available - check if this kernel supports PPP.
 */
int
ppp_available()
{
}

/*
 * establish_ppp - Turn the serial port into a ppp interface.
 */
void
establish_ppp()
{
}

/*
 * disestablish_ppp - Restore the serial port to normal operation.
 * It attempts to reconstruct the stream with the previously popped
 * modules.  This shouldn't call die() because it's called from die().
 */
void
disestablish_ppp()
{
}

/*
 * output - Output PPP packet.
 */
void
output(unit, p, len)
int unit;
u_char *p;
int len;
{
}

/*
 * read_packet - get a PPP packet from the serial device.
 */
int
read_packet(buf)
u_char *buf;
{
}
#endif

/*
 * ppp_send_config - configure the transmit characteristics of
 * the ppp interface.
 */
void
ppp_send_config(unit, mtu, asyncmap, pcomp, accomp, err)
int unit;
int mtu;
u_long asyncmap;
int pcomp;
int accomp;
int err;
{
    struct ifreq ifr;
    char c;

    strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));

    /* set MTU */

    if (mtu <= MTU) {
	ifr.ifr_mtu = mtu;
	if (ioctl(s, SIOCSIFMTU, (caddr_t) & ifr) < 0) {
	    if (!(err && err == errno)) {
		syslog(LOG_ERR, "ioctl(SIOCSIFMTU): %m");
		quit();
	    }
	}
    }

    /* set asyncmap */

    if (!syncmode && strioctl(dpfd, DPIOCSASYNCMAP, -1, sizeof(asyncmap),
	    (caddr_t) &asyncmap) < 0) {
	if (!(err && err == errno))
	    syslog(LOG_ERR, "ioctl(DPIOCSASYNCMAP): %m");
    }

    /* set protocol compression */

    c = pcomp;
    if (!syncmode && strioctl(dpfd, DPIOCSCOMPPROT, -1, sizeof(c), (caddr_t) & c) < 0) {
	if (!(err && err == errno))
	    syslog(LOG_ERR, "ioctl(DPIOCSCOMPPROT): %m");
    }

    /* set address/control compression */

    c = accomp;
    if (!syncmode && strioctl(dpfd, DPIOCSCOMPAC, -1, sizeof(c), (caddr_t) & c) < 0) {
	if (!(err && err == errno))
	    syslog(LOG_ERR, "ioctl(DPIOCSCOMPAC): %m");
    }
}


#if 0
/*
 * ppp_set_xaccm - set the extended transmit ACCM for the interface.
 */
void
ppp_set_xaccm(unit, accm)
int unit;
ext_accm accm;
{
}
#endif


/*
 * ppp_recv_config - configure the receive-side characteristics of
 * the ppp interface.
 */
void
ppp_recv_config(unit, mru, asyncmap, pcomp, accomp)
int unit;
int mru;
u_long asyncmap;
int pcomp;
int accomp;
{
    /* do nothing - dp doesn't distinguish tx and rx sides */
}

/*
 * sifframing - Config the interface for SLIP framing
 */
int
sifframing(u, slip)
int u;
int slip;
{
    u_char x = slip ? DP_FRAME_SLIP : DP_FRAME_PPP;

    if (debug)
        syslog(LOG_DEBUG, "SIFFRAMING unit %d to value %d\n", u, x);
    if (strioctl(dpfd, DPIOCSFRAMING, -1, sizeof(x), (caddr_t) & x) < 0) {
        syslog(LOG_ERR, "ioctl(DPIOCSFRAMING): %m");
        return 0;	/* XXX or quit() ? */
    }
    return 1;
}

/*
 * sifvjcomp - config tcp header compression
 */
int
sifvjcomp(u, vjcomp, cidcomp, maxcid)
int u;
int vjcomp;
int cidcomp;
int maxcid;
{
    if (vjcomp) {
	u_char x[2]; x[0] = maxcid; x[1] = cidcomp;
	if (debug)
	    syslog(LOG_DEBUG, "SIFVJCOMP unit %d to value %d.%d\n",
		u, maxcid, cidcomp);
	if (strioctl(dpfd, DPIOCSVJCOMP, -1, sizeof(x), (caddr_t)x) < 0) {
	    syslog(LOG_ERR, "ioctl(DPIOCSVJCOMP): %m");
	    return 0;	/* XXX or quit() ? */
	}
    }
    return 1;
}

/*
 * sifup - Config the interface up.
 */
int
sifup(u)
int u;
{
    struct ifreq ifr;

    if (dialup)
    	/* don't fiddle with ifr_flags - dp interfaces are always "up" */
	return 1;

    strncpy(ifr.ifr_name, ifname, sizeof (ifr.ifr_name));
    if (ioctl(s, SIOCGIFFLAGS, (caddr_t) &ifr) < 0) {
	syslog(LOG_DEBUG, "ioctl (SIOCGIFFLAGS): %m");
	return 0;
    }
    ifr.ifr_flags |= IFF_UP;
    if (ioctl(s, SIOCSIFFLAGS, (caddr_t) &ifr) < 0) {
	syslog(LOG_DEBUG, "ioctl (SIOCSIFFLAGS): %m");
	return 0;
    }
    return 1;
}

/*
 * sifdown - Config the interface down.
 */
int
sifdown(u)
int u;
{
    struct ifreq ifr;

    if (dialup)
    	/* don't fiddle with ifr_flags - dp interfaces are always "up" */
	return 1;

    strncpy(ifr.ifr_name, ifname, sizeof (ifr.ifr_name));
    if (ioctl(s, SIOCGIFFLAGS, (caddr_t) &ifr) < 0) {
	syslog(LOG_DEBUG, "ioctl (SIOCGIFFLAGS): %m");
	return 0;
    }
    ifr.ifr_flags &= ~IFF_UP;
    if (ioctl(s, SIOCSIFFLAGS, (caddr_t) &ifr) < 0) {
	syslog(LOG_DEBUG, "ioctl (SIOCSIFFLAGS): %m");
	return 0;
    }
    return 1;
}

/*
 * SET_SA_FAMILY - initialize a struct sockaddr, setting the sa_family field.
 */
#define SET_SA_FAMILY(addr, family)		\
    BZERO((char *) &(addr), sizeof(addr));	\
    addr.sa_family = (family);

/*
 * sifaddr - Config the interface IP addresses and netmask.
 */
int
sifaddr(u, o, h, m)
int u;
u_long o;
u_long h;
u_long m;
{
    struct ifreq ifr;
    int ret = 1;

    if (!dialup || dynamic) {
	strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
	(void) memset((char *) &ifr.ifr_addr, 0, sizeof(ifr.ifr_addr));
	ifr.ifr_addr.sa_family = AF_INET;
	((struct sockaddr_in *) & ifr.ifr_addr)->sin_addr.s_addr = o;
	if (ioctl(s, SIOCSIFADDR, (caddr_t) & ifr) < 0) {
	    syslog(LOG_ERR, "ioctl(SIOCSIFADDR): %m");
	    ret = 0;
	}
	((struct sockaddr_in *) & ifr.ifr_dstaddr)->sin_addr.s_addr = h;
	if (ioctl(s, SIOCSIFDSTADDR, (caddr_t) & ifr) < 0) {
	    syslog(LOG_ERR, "ioctl(SIOCSIFDSTADDR): %m");
	    ret = 0;
	}
	((struct sockaddr_in *) & ifr.ifr_addr)->sin_addr.s_addr = m;
	if (ioctl(s, SIOCSIFNETMASK, (caddr_t) & ifr) < 0) {
	    syslog(LOG_ERR, "ioctl(SIOCSIFNETMASK): %m");
	    ret = 0;
	}
    }
    return ret;
}

/*
 * cifaddr - Clear the interface IP addresses, and delete routes
 * through the interface if possible.
 */
int
cifaddr(u, o, h)
int u;
u_long o;
u_long h;
{
    /* do nothing for dp */
}

/*
 * sifbringup - Start packets flowing.
 */
int
sifbringup(s)
char s;
{
    char stat = s;
    if (strioctl(dpfd, DPIOCCALLSTAT, -1, sizeof(stat), &stat) < 0) {
	syslog(LOG_ERR, "ioctl (DPIOCCALLSTAT): %m");
	quit();
    }
    return 1;
}

/*
 * sifsapenable - Start packets flowing for a particular SAP (protocol)
 */
int
sifsapenable(s)
ulong s;
{
    if (dialup) {
	ulong sap = s;
	if (strioctl(dpfd, DPIOCSAPENABLE, -1, sizeof(sap), (caddr_t)&sap) < 0) {
	    syslog(LOG_ERR, "ioctl (DPIOCSAPENABLE): %m");
	    quit();
	}
    }
    return 1;
}

/*
 * sifsapdisable - Stop packets flowing for a particular SAP (protocol)
 */
int
sifsapdisable(s)
ulong s;
{
    if (dialup) {
	ulong sap = s;
	if (strioctl(dpfd, DPIOCSAPDISABLE, -1, sizeof(sap), (caddr_t)&sap) < 0) {
	    syslog(LOG_ERR, "ioctl (DPIOCSAPDISABLE): %m");
	    quit();
	}
    }
    return 1;
}
