/*
 * ppp_sync.c - Streams sync functions 
 *
 * Copyright (C) 1990  Brad K. Clements, All Rights Reserved
 * fcstab and some ideas nicked from if_ppp.c from cmu.
 * See copyright notice in if_ppp.h and NOTES
 *
 * $Id: pppsync.c,v 1.12 1995/09/22 13:53:41 ks Exp $
 *
 *
 * Copyright (c) 1993 Purdue University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Purdue University.  The name of the University may not be used
 * to endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Note: this copyright applies to portions of this software developed
 * at Purdue beyond the software covered by the original copyright.
 */


#include <sys/param.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/stream.h>
#include <sys/stropts.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>

#ifdef	SOLARIS_1
#include <sys/user.h>
#include <sys/kmem_alloc.h>
#else
#include <sys/kmem.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#endif

#if defined(SOLARIS_1) && !defined(LOADABLE)
#include <sys/ppp.h>
#include <sys/dpio.h>
#else
#include "ppp.h"
#include "dpio.h"
#endif

#include "args.h"

#ifdef	DEBUGS
#include <sys/syslog.h>
#define	static	
#ifdef	SOLARIS_1
#define	DLOG(s)		do{if(ppp_sync_debug)log(LOG_INFO,(s));}while(0)
#define	DLOG1(s, a)	do{if(ppp_sync_debug)log(LOG_INFO,(s),(a));}while(0)
#define	DLOG2(s, a, b)	do{if(ppp_sync_debug)log(LOG_INFO,(s),(a),(b));}while(0)
#else
#define	DLOG(s)		do{if(ppp_sync_debug)cmn_err(CE_CONT,(s));}while(0)
#define	DLOG1(s, a)	do{if(ppp_sync_debug)cmn_err(CE_CONT,(s),(a));}while(0)
#define	DLOG2(s, a, b)	do{if(ppp_sync_debug)cmn_err(CE_CONT,(s),(a),(b));}while(0)
#endif

int	ppp_sync_debug=0;
int	ppp_sync_input_debug=0;
#define MAX_DUMP_BYTES 1504
#else
#define	DLOG(s)
#define	DLOG1(s, a)
#define	DLOG2(s, a, b)
#endif

#ifdef	SOLARIS_1
/*
 * Define a number of facilities to allow the code to look more
 * like Solaris 2.X.
 */
#define D_MP 0
#define qprocson(x)
#define qprocsoff(x)
#define	mutex_enter(x)
#define	mutex_exit(x)
#define canputnext(q)   (canput((q)->q_next))
#define KM_SLEEP        KMEM_SLEEP
#define KM_NOSLEEP      KMEM_NOSLEEP
#define	const
#endif


static int ppp_sync_open	__ARGS((queue_t *, dev_t *, int, int,
					cred_t *));
static int ppp_sync_close	__ARGS((queue_t *, cred_t *));
static int ppp_sync_rput	__ARGS((queue_t *, mblk_t *));
static int ppp_sync_rsrv	__ARGS((queue_t *));
static int ppp_sync_wput	__ARGS((queue_t *, mblk_t *));
static int ppp_sync_wsrv	__ARGS((queue_t *));

static 	struct	module_info	ppp_sync_info ={
    0xabcd, "pppsync", 0, INFPSZ, 16384, 4096
};

static	struct	qinit	r_init = {
    ppp_sync_rput, ppp_sync_rsrv, ppp_sync_open, ppp_sync_close,
    NULL, &ppp_sync_info, NULL
};
static	struct	qinit	w_init = {
    ppp_sync_wput, ppp_sync_wsrv, ppp_sync_open, ppp_sync_close,
    NULL, &ppp_sync_info, NULL
};
struct	streamtab	pppsyncinfo = {
    &r_init, &w_init, NULL, NULL
};

typedef struct  ppp_sync_info {
    u_short pai_flags;		/* status flags */
    struct ppp_sync_info
	    *next_pai;		/* next PAI in active list */
} pai_t;

/*
 * Values for pai_flags
 */
#define	PAI_FLAGS_FLUSH		0x01
#define	PAI_FLAGS_COMPPROT	0x04
#define	PAI_FLAGS_COMPAC	0x08

static pai_t *pais;		/* Our list of async control structures */
#ifndef	SOLARIS_1
static kmutex_t pais_lock;
#endif

#ifdef	SOLARIS_1
static int
ppp_sync_open(rq, dev, flag, sflag)
queue_t *rq;
dev_t dev;
int flag,
    sflag;
#else
static int
ppp_sync_open(queue_t *rq, dev_t *devp, int flag, int sflag, cred_t *credp)
#endif
{
    register pai_t	*p;
    int	s;

    DLOG("ppp_sync_open\n");
    if (rq->q_ptr)
	return 0;

#ifdef	SOLARIS_1
    if (!suser()) {
	u.u_error = EPERM;
	return OPENFAIL;
    }
    if (rq->q_ptr) {
	u.u_error = EBUSY;
	return OPENFAIL;
    }
#endif
#ifdef	SOLARIS_1
    s = splstr();
#endif
    if ((p = (pai_t *)kmem_zalloc(sizeof(pai_t), KM_SLEEP)) == NULL) {
#ifdef	SOLARIS_1
	splx(s);
	u.u_error = ENOMEM;
	return OPENFAIL;
#else
	return ENOMEM;
#endif
    }

    rq->q_ptr = WR(rq)->q_ptr = (caddr_t)p;
    p->pai_flags = PAI_FLAGS_FLUSH;
    DLOG("ppp_sync: opening\n");
    mutex_enter(&pais_lock);
    p->next_pai = pais;
    pais = p;
    mutex_exit(&pais_lock);
    qprocson(rq);
#ifdef	SOLARIS_1
    splx(s);
#endif
    return(0);
}

#ifdef	SOLARIS_1
static int
ppp_sync_close(rq)
queue_t *rq;
#else
static int
ppp_sync_close(queue_t *rq, cred_t *credp)
#endif
{
    int	s;
    register pai_t 	*p, **pp;

    DLOG("ppp_sync_close\n");
#ifdef	SOLARIS_1
    s = splstr();
#endif
    qprocsoff(rq);

    mutex_enter(&pais_lock);

    for (pp = &pais ; p = *pp ; pp = &p->next_pai)
	if (p == (pai_t *)rq->q_ptr)
	    break;
    *pp = p->next_pai;

    mutex_exit(&pais_lock);

    /*
     * Free associated memory.
     */
    kmem_free((caddr_t)p, sizeof(pai_t));

    rq->q_ptr = WR(rq)->q_ptr = (caddr_t)0;


    DLOG("ppp_sync: closing\n");
#ifdef	SOLARIS_1
    splx(s);
#endif
    return(0);			
}


/*
 * M_IOCTL processing is performed at this level. There is some 
 * weirdness here, but I couldn't think of an easier way to handle it.
 *
 * DPIOCSASYNCMAP and DPIOCGASYNCMAP is handled here.
 *
 * DPIOCSCOMPAC and DPIOCSCOMPPROT are both handled here, rather
 * than jamming new flag bits into the if_ interface.  However, older
 * versions of the upper modules may note these ioctls as they go by.
 *
 * DPIOCSMRU and DPIOCGMRU (Max Receive Unit) are both handled here.
 * Rather than using the MTU to set the MRU, we have a seperate IOCTL for it.
 *
 */
static int
ppp_sync_wput(q, mp)
queue_t  *q;
register mblk_t *mp;
{
    switch (mp->b_datap->db_type) {
     case M_FLUSH:
	if (*mp->b_rptr & FLUSHW)
	    flushq(q, FLUSHDATA);
	putnext(q, mp);			/* send it along too */
	break;

      case M_DATA:
	putq(q, mp);			/* queue it for my service routine */
	break;

      case M_IOCTL:
	ppp_sync_ioc(q, mp);
	break;

     default:
	putnext(q,mp);	/* don't know what to do with this, so send it along*/
    }
}

#ifdef	SOLARIS_1
#ifndef	TRANSPARENT
#define	TRANSPARENT (unsigned int)(-1)
#endif
#endif

static int
ppp_sync_ioc(q, mp)
queue_t *q;
mblk_t *mp;
{
    register struct iocblk	*ioc;
    register pai_t	*p;
    int	x, flags;
    int err = 0;

    ioc = (struct iocblk *) mp->b_rptr;
    p =  (pai_t *) q->q_ptr;

    if (ioc->ioc_count == TRANSPARENT) {
	/*
	 * No TRANSPARENT ioctl support.
	 */
	DLOG1("transparent ioctl: %x\n", ioc->ioc_cmd);
	if (mp->b_cont)
	    DLOG2("transparent ioctl data?: 0x%x %x\n",
		  mp->b_cont->b_datap->db_type, msgdsize(mp->b_cont));
	err = EINVAL;
	goto pa_ioc_ret;
    }
    switch (ioc->ioc_cmd) {
     case DPIOCSCOMPAC:	/* enable or disable AC compression */
	if (ioc->ioc_count != sizeof(u_char)) {
	    err = EINVAL;
	    break;
	}
	DLOG1("ppp_sync: SCOMPAC %d\n", *(u_char *)mp->b_cont->b_rptr);
	if ( *(u_char *) mp->b_cont->b_rptr) 
	    p->pai_flags |= PAI_FLAGS_COMPAC;
	else
	    p->pai_flags &= ~PAI_FLAGS_COMPAC;
	break;

     case DPIOCSCOMPPROT:	/* enable or disable PROT  compression */
	if (ioc->ioc_count != sizeof(u_char)) {
	    err = EINVAL;
	    break;
	}
	DLOG1("ppp_sync: SCOMPPROT %d\n", *(u_char *)mp->b_cont->b_rptr);
	if ( *(u_char *) mp->b_cont->b_rptr) 
	    p->pai_flags |= PAI_FLAGS_COMPPROT;
	else
	    p->pai_flags &= ~PAI_FLAGS_COMPPROT;
	break;

     case DPIOCSMRU:
	if (ioc->ioc_count != sizeof(int)) {
	    err = EINVAL;
	    break;
	}
	x = *(int *) mp->b_cont->b_rptr;
	if (x < PPP_MTU)
	    x = PPP_MTU;
	x += PPP_HDR_SIZE + sizeof(u_short) + 1;
	if (x > 4096) {		/* couldn't allocb something this big */
	    err = EINVAL;
	    break;
	}
	break;

     case DPIOCGMRU:
	if (ioc->ioc_count != sizeof(int)) {
	    err = EINVAL;
	    break;
	}
	*(int *)mp->b_cont->b_wptr = PPP_MTU;
/*		p->pai_buffsize - (PPP_HDR_SIZE + sizeof(u_short) + 1);*/
	goto pa_ioc_data_ret;

     case DPIOCGASYNCMAP:
	if (ioc->ioc_count != sizeof(u_long)) {
	    err = EINVAL;
	    break;
	}
	*(u_long *) mp->b_cont->b_wptr = 0;
	goto pa_ioc_data_ret;

     case DPIOCSASYNCMAP:
	if (ioc->ioc_count != sizeof(u_long)) {
	    err = EINVAL;
	    break;
	}
	DLOG1("ppp_sync: SASYNCMAP %x\n",
	      *(u_long *)mp->b_cont->b_rptr);

	break;

     case DPIOCGASDEBUG :
	if (ioc->ioc_count != sizeof(int)) {
	    err = EINVAL;
	    break;
	}
#ifdef	DEBUGS
	*(int *)mp->b_cont->b_wptr = (ppp_sync_debug ? 1 : 0)
				   | (ppp_sync_input_debug ? 2 : 0);
#endif
	goto pa_ioc_data_ret;

     case DPIOCSASDEBUG :
	if (ioc->ioc_count != sizeof(int)) {
	    err = EINVAL;
	    break;
	}
	flags = *(int *)mp->b_cont->b_rptr;
	DLOG1("ppp_sync: SASDEBUG %x\n", flags);

#ifdef	DEBUGS
	ppp_sync_debug = flags & 1;
	ppp_sync_input_debug = flags & 2;
#endif
	break;

     default:
	putnext(q, mp);
	return;
    }

pa_ioc_ret:
    if (mp->b_cont) {
	freemsg(mp->b_cont);
	mp->b_cont = NULL;
    }
    ioc->ioc_count = 0;
pa_ioc_data_ret:
    if (err) {
	ioc->ioc_error = err;
	mp->b_datap->db_type = M_IOCNAK;
    }
    else
	mp->b_datap->db_type = M_IOCACK;
    qreply(q, mp);
}

/*
 * This is an incredibly ugly routine. If you see a better way of
 * doing this, feel free to improve it. I'm hoping that the buffer
 * management routines are efficient, in terms of dup'ing and adjusting
 * message blocks. Hopefully we won't run out of small message blocks.
 * Perhaps some counters should be kept to record how efficient this
 * routine is, and how much break up of messages is required.  I'm not
 * doing anything funny with stuffing PPP_ESCAPES into the actual data
 * buffers.  If you get a lot of outgoing errors (under Solaris 1.X)..
 * you might need to up your NBLK4 parameter  by quite a bit.... (see param.c)
 */
static int
ppp_sync_wsrv(q)
queue_t	*q;
{
    register pai_t	*p = (pai_t *)q->q_ptr;
    register mblk_t	*mp;

    while ((mp = getq(q)) != NULL) {
	/*
	 * we can only get M_DATA types into our Queue,
	 * due to our Put function
	 */
	if (!canputnext(q)) {
	    putbq(q, mp);
	    return;
	}

	/*
	 * at least a ppp header required
	 */
	if (msgdsize(mp) < PPP_HDR_SIZE) {	
	    freemsg(mp);		/* discard the message */
	    putctl1(OTHERQ(q), M_CTL, DP_IF_OERROR); /* indicate output err */
	    continue;
	}

	/*
	 * Compress PPP header, if allowed.
	 */
	if ((p->pai_flags & PAI_FLAGS_COMPAC) &&
	    *mp->b_rptr == PPP_ALLSTATIONS &&
	    *(mp->b_rptr+1) == PPP_UI)
	    mp->b_rptr += 2;
	if ((p->pai_flags & PAI_FLAGS_COMPPROT) &&
	    *mp->b_rptr == 0)
	    mp->b_rptr++;

	/* phew, ready to ship */
	putnext(q, mp);
    } /* end while (getq()) */
}

static int
ppp_sync_rput(q, mp)
queue_t *q;
register mblk_t *mp;
{
    switch (mp->b_datap->db_type) {

     case M_FLUSH:
	if (*mp->b_rptr & FLUSHR)
	    flushq(q, FLUSHDATA);
	putnext(q, mp);			/* send it along too */
	break;

     case M_DATA:
	putq(q, mp);			/* queue it for my service routine */
	break;

     default:
	putnext(q,mp);	/* don't know what to do with this, so send it along */
    }
}

#ifdef	DEBUGS
static const char ppp_digits[] = "0123456789abcdef";
#endif

#define	INPUT_ERROR(q)	putctl1(q, M_CTL, DP_IF_IERROR)

static int
ppp_sync_rsrv(q)
queue_t	*q;
{
    register mblk_t	*mp;
    register u_char	*cp,c;
    int			mleft;
    register pai_t	*p;
    mblk_t		*m0;

    p = (pai_t *) q->q_ptr;

    while ((mp = getq(q)) != NULL) {
	phdr_t *ph;
	/*
	 * Our put function limits messages to type M_DATA.
	 */
	if (!canputnext(q)) {
	    putbq(q, mp);
	    return;
	}

	if (!(m0 = allocb(PPP_HDR_SIZE, BPRI_MED))) {
		DLOG("ppp_sync_rsrv: can't alloc PPP header\n");
	    	putbq(q, mp);
	    	return;
	}
	ph = (phdr_t *)m0->b_wptr;
	m0->b_wptr += PPP_HDR_SIZE;
	ph->ph_address = PPP_ALLSTATIONS;
	ph->ph_control = PPP_UI;
	pullupmsg(mp, 4); /* XXX necessary? */
	mleft = msgdsize(mp);
	if((mleft >= 2) &&
	    (*mp->b_rptr == PPP_ALLSTATIONS) && (*(mp->b_rptr+1) == PPP_UI)) {
		mp->b_rptr += 2;
		mleft -= 2;
	}

	if(mleft < 1)
		goto tooshort;

	if ((*(mp->b_rptr) & 1) == 1) {
		if (p->pai_flags & PAI_FLAGS_COMPPROT) {
			ph->ph_protocol = *mp->b_rptr++;
			mleft--;
		}
		else {
			DLOG("ppp_sync_rsrv: bad protocol high byte\n");
			goto rerr;
		}
	} else {
		if(mleft < 2)
			goto tooshort;
		if ((*(mp->b_rptr+1) & 1) != 1) {
			DLOG("ppp_sync_rsrv: bad protocol low byte\n");
			goto rerr;
		}
		ph->ph_protocol = *((ushort *)mp->b_rptr);
		mp->b_rptr += 2;
		mleft -= 2;
	}

	linkb(m0, mp);
	putnext(q, m0);
	continue;

tooshort:
	DLOG("ppp_sync_rsrv: short message\n");
rerr:
	freemsg(m0);
	freemsg(mp);
	INPUT_ERROR(q);
    }	/* end while  getq */
}

static char pppsync_name[] = "PPP Sync v4.0";

#if defined(SOLARIS_1) && defined(LOADABLE)
#include <sys/conf.h>
#include <sys/buf.h>
#include <sundev/mbvar.h>
#include <sun/autoconf.h>
#include <sun/vddrv.h>

static struct vdldrv ppp_sync_vd = {
    VDMAGIC_PSEUDO,
    pppsync_name,
};
static struct fmodsw *fmod_ppp_sync;

xxxinit(fc,vdp,vdi,vds)
unsigned int fc;
struct vddrv *vdp;
addr_t vdi;
struct vdstat *vds;
{

    switch (fc) {
     case VDLOAD:
	{
	    int dev,i;
	    for (dev = 0; dev < fmodcnt; dev++) {
		if (fmodsw[dev].f_str == NULL)
		    break;
	    }
	    if (dev == fmodcnt)
		return ENODEV;
	    fmod_ppp_sync = &fmodsw[dev];
	    for (i = 0; i <= FMNAMESZ; i++)
		fmod_ppp_sync->f_name[i] = ppp_sync_info.mi_idname[i];
	    fmod_ppp_sync->f_str = &pppsyncinfo;
	    vdp->vdd_vdtab = (struct vdlinkage *) &ppp_sync_vd;
	}
	return 0;
     case VDUNLOAD:
	if (pais)
	    return EBUSY;

	fmod_ppp_sync->f_name[0] = '\0';
	fmod_ppp_sync->f_str = NULL;
	return 0;
     case VDSTAT:
	return 0;
     default:
	return EIO;
    }
}
#elif !defined(SOLARIS_1)
/*
 * Solaris 2.X version
 */
#include <sys/modctl.h>

extern struct streamtab pppsyncinfo;

static struct fmodsw pppsyncfsw = {
    "pppsync",
    &pppsyncinfo,
    D_NEW | D_MP
};
	
extern struct mod_ops mod_strmodops;

static struct modlstrmod pppsynclstrmod = {
    &mod_strmodops,
    pppsync_name,
    &pppsyncfsw
};

static struct modlinkage pppsyncmodlinkage = {
    MODREV_1,
    (void *)&pppsynclstrmod,
    NULL
};

/*
 * Module Config Entry Points.
 */

_init(void)
{
    mutex_init(&pais_lock, "pppsync module list lock",
	       MUTEX_DRIVER, (void *)0);
    return (mod_install(&pppsyncmodlinkage));
}

_fini(void)
{
    register int ret;

    if ((ret = mod_remove(&pppsyncmodlinkage)) == DDI_SUCCESS) {
	mutex_destroy(&pais_lock);
	return ret;
    }
    else
	return EBUSY;
}

_info(struct modinfo *modinfop)
{
    return mod_info(&pppsyncmodlinkage, modinfop);
}
#endif
